#ifndef LANDER_HTTP_LOOP
#define LANDER_HTTP_LOOP

#include <regex.h>

#include "event_loop.h"
#include "http/req.h"
#include "http/res.h"
#include "http/types.h"
#include "trie.h"

// Max amount of steps a route can use
#define HTTP_LOOP_MAX_STEPS 17

#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#define MAX(x, y) (((x) > (y)) ? (x) : (y))

/**
 * Type of a route
 */
typedef enum http_route_type {
  http_route_literal = 0,
  http_route_regex = 1,
} http_route_type;

/**
 * Function describing a step in a route's processing.
 *
 * @param conn connection to process
 * @return whether the processing can immediately advance to the next step. A
 * step should return false if it's e.g. waiting for I/O, and can therefore not
 * finish its task in the current cycle of the event loop.
 */
typedef bool (*step)(event_loop_conn *conn);

/**
 * Struct describing a route a request can take.
 */
typedef struct http_route {
  http_route_type type;
  http_method method;
  char *path;
  // Compiled regex for a regex route. This value gets set at runtime when
  // starting the http loop
  regex_t *regex;
  step steps[HTTP_LOOP_MAX_STEPS];
} http_route;

/**
 * Global context passed to every connection using the same pointer
 */
typedef struct http_loop_gctx {
  http_route *routes;
  size_t route_count;
  Trie *trie;
  const char *api_key;
  const char *data_dir;
} http_loop_gctx;

/**
 * Initialize a new global context
 *
 * @return pointer to the newly allocated object.
 */
http_loop_gctx *http_loop_gctx_init();

/**
 * Invidivual context initialized for every connection
 */
typedef struct http_loop_ctx {
  http_request req;
  http_response res;
  http_route *route;
  size_t current_step;
  http_loop_gctx *g;
} http_loop_ctx;

/**
 * Initialize a context struct
 *
 * @param g global context
 * @return pointer to the newly allocated context
 */
http_loop_ctx *http_loop_ctx_init(http_loop_gctx *g);

/**
 * Resets an already allocated context so that it can be reused for a new
 * request.
 *
 * @param ctx context to reset
 */
void http_loop_ctx_reset(http_loop_ctx *ctx);

/**
 * Free a context struct. Internally this first calls http_loop_ctx_reset.
 *
 * @param ctx context to free
 */
void http_loop_ctx_free(http_loop_ctx *ctx);

/**
 * Process incoming data as an HTTP request. This is the "handle_data" function
 * for the event loop.
 *
 * @param conn connection to process
 * @return whether another request can be processed immediately.
 */
bool http_loop_handle_request(event_loop_conn *conn);

/**
 * Write the HTTP response to the file descriptor. This is the "write_data"
 * function for the event loop.
 *
 * @param conn connection to process
 */
void http_loop_write_response(event_loop_conn *conn);

/**
 * Try to parse the incoming data as an HTTP request.
 *
 * @param conn connection to process
 * @return result of the parse
 */
http_parse_error http_loop_parse_request(event_loop_conn *conn);

/**
 * Try to match the parsed request with one of the defined routes, aka route the
 * request.
 *
 * @param conn connection to process
 */
void http_loop_route_request(event_loop_conn *conn);

/**
 * Advance the processing of the routed request's processing by cycling through
 * the request's various steps.
 *
 * @param conn connection to process
 */
void http_loop_process_request(event_loop_conn *conn);

/**
 * Request step that consumes the request body and stores it in a buffer.
 *
 * @param conn connection to process
 * @return true if the body has been fully received, false otherwise
 */
bool http_loop_step_body_to_buf(event_loop_conn *conn);

/**
 * Request step that consumes the request body and stores it in a file.
 *
 * @param conn connection to process
 * @return true if the body has been fully received, false otherwise
 */
bool http_loop_step_body_to_file(event_loop_conn *conn);

/**
 * Authenticate the request using the X-Api-Key header.
 *
 * @param conn connection to check
 * @return always true
 */
bool http_loop_step_auth(event_loop_conn *conn);

/**
 * A step that simply sets the connection's state to res.
 *
 * @param conn connection to process
 * @return always true
 */
bool http_loop_step_switch_res(event_loop_conn *conn);

/**
 * Initialize a new http loop.
 *
 * @param gctx global context for the event loop
 * @return pointer to the newly allocated object
 */
event_loop *http_loop_init(http_loop_gctx *gctx);

/**
 * Run the HTTP loop. This function never returns.
 *
 * @param el the event loop
 * @param port on what port to listen
 */
void http_loop_run(event_loop *el, int port);

#endif
